import requests
from requests.exceptions import HTTPError

from anylearn.sdk.auth import configure_auth
from anylearn.utils.retry import RequestRetrySession


def raise_for_status(res: requests.Response):
    http_error_msg = ''
    if isinstance(res.reason, bytes):
        try:
            reason = res.reason.decode('utf-8')
        except UnicodeDecodeError:
            reason = res.reason.decode('iso-8859-1')
    else:
        reason = res.reason
    
    if isinstance(res.content, bytes):
        try:
            content = res.content.decode('utf-8')
        except UnicodeDecodeError:
            content = res.content.decode('iso-8859-1')
    else:
        content = res.content

    if 400 <= res.status_code < 500:
        http_error_msg = u'%s: "%s" for url: %s' % (res.status_code, content, res.url)

    elif 500 <= res.status_code < 600:
        http_error_msg = u'%s: "%s" for url: %s' % (res.status_code, reason, res.url)

    if http_error_msg:
        raise HTTPError(http_error_msg, response=res)


def request(method, *args, **kwargs):
    with RequestRetrySession() as sess:
        res = sess.request(method, *args, **kwargs)
        raise_for_status(res)
        res.encoding = "utf-8"
        return res.json()


def request_with_token(method, *args, **kwargs):
    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    auth = configure_auth()
    kwargs['headers']['Authorization'] = f"Bearer {auth.access_token}"
    try:
        return request(method, *args, **kwargs)
    except HTTPError as e:
        if e.response.status_code == 401 and auth.refresh_token:
            auth = configure_auth(force=True)
            if auth:
                return request_with_token(method, *args, **kwargs)
        raise


def get_with_token(*args, **kwargs):
    return request_with_token('GET', *args, **kwargs)


def post_with_token(*args, **kwargs):
    return request_with_token('POST', *args, **kwargs)


def delete_with_token(*args, **kwargs):
    return request_with_token('DELETE', *args, **kwargs)


def put_with_token(*args, **kwargs):
    return request_with_token('PUT', *args, **kwargs)


def patch_with_token(*args, **kwargs):
    return request_with_token('PATCH', *args, **kwargs)



import requests
from urllib3.util.retry import Retry
from requests.exceptions import HTTPError
from requests.adapters import HTTPAdapter

from anylearn.config import AnylearnConfig


url_base = lambda :AnylearnConfig.cluster_address + '/api'


class RequestRetrySession(object):
    def __init__(
        self,
        retries=3,
        backoff_factor=1,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["HEAD", "GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "TRACE"],
        session=None,
    ):
        self.session = session or requests.session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            allowed_methods=allowed_methods,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)

    def __enter__(self):
        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()


def raise_for_status(res: requests.Response):
    http_error_msg = ''
    if isinstance(res.reason, bytes):
        try:
            reason = res.reason.decode('utf-8')
        except UnicodeDecodeError:
            reason = res.reason.decode('iso-8859-1')
    else:
        reason = res.reason
    
    if isinstance(res.content, bytes):
        try:
            content = res.content.decode('utf-8')
        except UnicodeDecodeError:
            content = res.content.decode('iso-8859-1')
    else:
        content = res.content

    if 400 <= res.status_code < 500:
        http_error_msg = u'%s: "%s" for url: %s' % (res.status_code, content, res.url)

    elif 500 <= res.status_code < 600:
        http_error_msg = u'%s: "%s" for url: %s' % (res.status_code, reason, res.url)

    if http_error_msg:
        raise HTTPError(http_error_msg, response=res)


def __request(method, *args, **kwargs):
    with RequestRetrySession() as sess:
        res = sess.request(method, *args, **kwargs)
        raise_for_status(res)
        res.encoding = "utf-8"
        return res.json()


def __request_with_token(method, *args, **kwargs):
    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    kwargs['headers']['Authorization'] = "Bearer %s" % AnylearnConfig.token
    try:
        return __request(method, *args, **kwargs)
    except HTTPError as e:
        if e.response.status_code == 401 and AnylearnConfig.refresh_token:
            AnylearnConfig.cluster_relogin_by_token()
            return __request_with_token(method, *args, **kwargs)
        else:
            raise


def __request_with_secret_key(method, *args, secret_key, **kwargs):
    if 'headers' not in kwargs:
        kwargs['headers'] = {}
    kwargs['headers']['Secret-Key'] = secret_key
    return __request(method, *args, **kwargs)


def get_with_token(*args, **kwargs):
    return __request_with_token('GET', *args, **kwargs)


def post_with_token(*args, **kwargs):
    return __request_with_token('POST', *args, **kwargs)


def post_with_secret_key(*args, secret_key, **kwargs):
    return __request_with_secret_key('POST', *args, secret_key=secret_key, **kwargs)


def delete_with_token(*args, **kwargs):
    return __request_with_token('DELETE', *args, **kwargs)


def put_with_token(*args, **kwargs):
    return __request_with_token('PUT', *args, **kwargs)


def patch_with_token(*args, **kwargs):
    return __request_with_token('PATCH', *args, **kwargs)



import random
from typing import Optional

from anylearn.interfaces.base import BaseObject
from anylearn.utils.api import (
    url_base,
    patch_with_token,
)
from anylearn.utils.errors import AnyLearnException


TAG_COLOR = [
    { 'text_color': '#F53F3F', 'color': '#FCE7E7' },
    { 'text_color': '#FE8041', 'color': '#FFF3ED' },
    { 'text_color': '#FEB841', 'color': '#FCF7E1' },
    { 'text_color': '#1FC472', 'color': '#E9F9EF' },
    { 'text_color': '#35CBFC', 'color': '#E9F6F9' },
    { 'text_color': '#357FFC', 'color': '#EBF2FF' },
    { 'text_color': '#605AFF', 'color': '#EDEDFF' },
    { 'text_color': '#FFFFFF', 'color': '#F53F3F' },
    { 'text_color': '#FFFFFF', 'color': '#FE8041' },
    { 'text_color': '#FFFFFF', 'color': '#FEB841' },
    { 'text_color': '#FFFFFF', 'color': '#1FC472' },
    { 'text_color': '#FFFFFF', 'color': '#37BBE8' },
    { 'text_color': '#FFFFFF', 'color': '#357FFC' },
    { 'text_color': '#FFFFFF', 'color': '#605AFF' },
]


class TrainTag(BaseObject):
    """
    Anylearn训练任务标签类，仅支持quick_train打标签

    Attributes
    ----------
    id
        训练任务标签的唯一标识符，自动生成，由TTAG+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        标签的名称
    project_id
        所属项目的id
    """

    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'color', 'text_color', 'project_id'],
            'update': ['id', 'name', 'color', 'text_color', 'project_id'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'color', 'text_color', 'project_id', 'description'],
            'update': ['id', 'name', 'color', 'text_color', 'project_id', 'description'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 color: Optional[str]=None,
                 text_color: Optional[str]=None,
                 project_id: Optional[str]=None,
                 description: Optional[str]=None):
        self.id = id
        self.project_id = project_id
        self.name = name
        self.color = color
        self.text_color = text_color
        self.description = description
    
    def _url_create(self):
        return f"{url_base()}/{self._namespace()}/tag"
    
    def _url_update(self):
        return f"{url_base()}/{self._namespace()}/tag"
    
    def _update(self):
        data = self._payload_update()
        res = patch_with_token(self._url_update(), data=data)
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        return self.id == res['data']
    
    def get_list(self):
        pass

    def get_detail(self):
        pass
        
    @staticmethod
    def generate_random_tag_color():
        color_item = TAG_COLOR[random.randint(0, len(TAG_COLOR)-1)]
        return color_item['color'], color_item['text_color']

    def _namespace(self):
        return "train_task"



class AnylearnError(Exception):
    """Base Anylearn error."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class AnylearnInvalidResponseError(AnylearnError):
    """Raised when the response is invalid."""


class AnylearnArtifactDuplicationError(AnylearnError):
    """Raised when an artifact is duplicated."""


class AnylearnArtifactTooLargeError(AnylearnError):
    """Raised when an artifact is too large."""



from abc import ABC, abstractmethod

import requests
from tqdm import tqdm

from anylearn.config import AnylearnConfig
from anylearn.utils import logger
from anylearn.utils.api import url_base


class ResourceUploader(ABC):
    """
    资源上传工具接口
    """

    @abstractmethod
    def run(self, resource_id: str, chunks: list):
        """
        执行资源上传,自定义资源上传需实现此方法

        Parameters
        ----------
        resource_id
            资源ID
        chunks
            被切割后的文件内容列表
        """
        raise NotImplementedError


class SyncResourceUploader(ResourceUploader):
    def run(self, resource_id: str, chunks: list):
        for i, chunk in enumerate(tqdm(chunks)):
            url = f"{url_base()}/resource/upload"
            headers = {'Authorization': f"Bearer {AnylearnConfig.token}"}
            files = {'file': chunk}
            data = {'file_id': resource_id, 'chunk': str(i)}
            res = requests.post(url, headers=headers, files=files, data=data)
            res.raise_for_status()



from pathlib import Path

from .func import *
from .logger import logger


DEFAULT_ANYLEARN_HOST = "https://anylearn.nelbds.cn"


def get_config_path() -> Path:
    config_dir = Path.home() / ".anylearn"
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir / "config.yaml"


def no_none_filter(data: dict):
    return dict(list(filter(lambda item:item[1] is not None, list(data.items()))))



import os
from pathlib import Path
from typing import Optional, Tuple

import typer
from requests import HTTPError
from rich import print
from typing_extensions import Annotated, Literal

from anylearn.cli._utils import HostOption
from anylearn.sdk.artifacts.artifact import Artifact, ArtifactState
from anylearn.sdk.artifacts.dataset import DatasetArtifact
from anylearn.sdk.artifacts.model import ModelArtifact
from anylearn.sdk.auth import authenticate
from anylearn.sdk.console import (
    console_error,
    console_success,
    console_warning,
)
from anylearn.sdk.context import init
from anylearn.sdk.errors import (
    AnylearnArtifactDuplicationError,
    AnylearnArtifactTooLargeError,
    AnylearnInvalidResponseError,
)
from anylearn.sdk.jumps.channel import JumpsChannel
from anylearn.sdk.jumps.uploader import JumpsUploader


app = typer.Typer()


@app.command()
def dataset(
    name: Annotated[str, typer.Argument(
        help="The name of the dataset to create.",
    )],
    path: Annotated[Path, typer.Argument(
        help="The local path (file or directory) of the dataset to upload.",
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        resolve_path=True,
    )],
    compress: Annotated[bool, typer.Option(
        help="Compress data during transfer (ex. rsync -z).",
    )] = True,
    host: str = HostOption,
):
    _jupload_artifact(
        artifact_type="dataset",
        name=name,
        path=path,
        compress=compress,
        host=host,
    )


@app.command()
def model(
    name: Annotated[str, typer.Argument(
        help="The name of the model to create.",
    )],
    path: Annotated[Path, typer.Argument(
        help="The local path (file or directory) of the model to upload.",
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        resolve_path=True,
    )],
    compress: Annotated[bool, typer.Option(
        help="Compress data during transfer (ex. rsync -z).",
    )] = True,
    host: str = HostOption,
):
    _jupload_artifact(
        artifact_type="model",
        name=name,
        path=path,
        compress=compress,
        host=host,
    )


def _jupload_artifact(
    artifact_type: Literal["dataset", "model"],
    name: str,
    path: os.PathLike,
    compress: bool,
    host: str,
) -> None:
    init(host)
    if not authenticate(host):
        raise typer.Abort()
    artifact, resume = _get_or_create_artifact(
        artifact_type=artifact_type,
        name=name,
    )
    _upload_via_jumps_channel(
        artifact_id=artifact.id,
        path=path,
        compress=compress,
        resume=resume,
    )
    console_success("Upload OK")


def _get_or_create_artifact(
    artifact_type: Literal["dataset", "model"],
    name: str,
) -> Tuple[Artifact, bool]:
    if artifact_type == "dataset":
        return _get_or_create_dataset(name)
    elif artifact_type == "model":
        return _get_or_create_model(name)


def _get_or_create_dataset(name: str) -> Tuple[DatasetArtifact, bool]:
    try:
        dataset = DatasetArtifact.from_full_name(name)
        if dataset.state != ArtifactState.CREATED:
            console_error(
                f"Dataset {name} "
                "is being uploaded via Web UI or "
                "is already fully uploaded."
            )
            raise typer.Abort()
        console_warning(f"Dataset {name} already exists in your namespace.")
        resume = typer.confirm("Do you want to resume previous upload?")
        if not resume:
            typer.confirm("Do you want to restart the upload?", abort=True)
        return dataset, resume
    except HTTPError:
        return _create_dataset(name), False


def _create_dataset(name: str) -> DatasetArtifact:
    try:
        print(f"Creating dataset {name}...")
        return DatasetArtifact.create(name=name)
    except AnylearnArtifactDuplicationError:
        console_error(f"Dataset {name} already exists in your namespace.")
        raise typer.Abort()
    except AnylearnInvalidResponseError:
        console_error(f"Failed to create dataset {name}.")
        raise typer.Abort()
    except Exception as e:
        console_error(f"An error occurred during dataset creation: {e}")
        raise typer.Abort()


def _get_or_create_model(name: str) -> Tuple[ModelArtifact, bool]:
    try:
        model = ModelArtifact.from_full_name(name)
        if model.state != ArtifactState.CREATED:
            console_error(
                f"Model {name} "
                "is being uploaded via Web UI or "
                "is already fully uploaded."
            )
            raise typer.Abort()
        console_warning(f"Model {name} already exists in your namespace.")
        resume = typer.confirm("Do you want to resume previous upload?")
        if not resume:
            typer.confirm("Do you want to restart the upload?", abort=True)
        return model, resume
    except HTTPError:
        return _create_model(name), False


def _create_model(name: str) -> ModelArtifact:
    try:
        print(f"Creating model {name}...")
        return ModelArtifact.create(name=name)
    except AnylearnArtifactDuplicationError:
        console_error(f"Model {name} already exists in your namespace.")
        raise typer.Abort()
    except AnylearnInvalidResponseError:
        console_error(f"Failed to create model {name}.")
        raise typer.Abort()
    except Exception as e:
        console_error(f"An error occurred during model creation: {e}")
        raise typer.Abort()


def _upload_via_jumps_channel(
    artifact_id: str,
    path: os.PathLike,
    compress: bool,
    resume: bool,
) -> None:
    jc = _get_jumps_channel_by_artifact_id(artifact_id, missing_ok=True)
    client_identifier = JumpsChannel.get_client_identifier(path)
    if resume and jc is not None and jc.client_code != client_identifier:
        console_error(
            "Local device or local artifact path "
            "has changed since previous upload."
        )
        raise typer.Abort()
    elif not resume or jc is None:
        jc = _create_jumps_channel(
            artifact_id=artifact_id,
            artifact_local_path=path,
        )
    print(f"Uploading {path} to jump server...")
    uploader = JumpsUploader(
        channel=jc,
        local_path=path,
        compress=compress,
    )
    if uploader.upload() != 0:
        console_error("Upload Failed (jump)")
        raise typer.Abort()
    print("Transforming into Anylearn asset...")
    if not jc.transform():
        console_error("Upload Failed (transform)")
        raise typer.Abort()
    jc.finish()


def _get_jumps_channel_by_artifact_id(
    artifact_id: str,
    missing_ok: bool = True,
) -> Optional[JumpsChannel]:
    try:
        print("Try to restore existing jump server channel...")
        return JumpsChannel.from_artifact_id(artifact_id)
    except:
        if not missing_ok:
            console_error(
                f"Jumps channel for artifact {artifact_id} not found."
            )
            raise typer.Abort()
        return None


def _create_jumps_channel(
    artifact_id: str,
    artifact_local_path: os.PathLike,
) -> JumpsChannel:
    try:
        print("Creating jump server channel...")
        return JumpsChannel.create(
            artifact_id=artifact_id,
            artifact_local_path=artifact_local_path,
        )
    except FileNotFoundError:
        console_error(f"Local path {artifact_local_path} not found.")
        raise typer.Abort()
    except AnylearnArtifactTooLargeError:
        raise typer.Abort()
    except AnylearnInvalidResponseError:
        console_error("Failed to create jumps channel.")
        raise typer.Abort()
    except Exception as e:
        console_error(f"An error occurred during jumps upload: {e}")
        raise typer.Abort()



import typer
from rich import print

import anylearn.cli.jupload as jupload
from anylearn.cli._utils import HostArgument
from anylearn.sdk.auth import authenticate, disauthenticate


app = typer.Typer(name="anyctl")
app.add_typer(jupload.app, name="jupload")


@app.command()
def login(host: str = HostArgument):
    if authenticate(host) is not None:
        print("[green]Login OK[/green]")
    else:
        print("[red]Login Failed[/red]")


@app.command()
def logout(host: str = HostArgument):
    disauthenticate(host)
    print("[green]Logout OK[/green]")



from dataclasses import dataclass

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa


@dataclass
class SSHKey:
    public_key: bytes
    private_key: bytes


RSA_PUBLIC_EXPONENT = 65537
RSA_KEY_SIZE = 4096


def generate_ssh_key_pair() -> SSHKey:
    # ref: https://stackoverflow.com/a/39126754/1720770
    key = rsa.generate_private_key(
        public_exponent=RSA_PUBLIC_EXPONENT,
        key_size=RSA_KEY_SIZE,
    )
    public_key = key.public_key().public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    )
    private_key = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return SSHKey(public_key=public_key, private_key=private_key)



from rich import print


def console_error(message: str) -> None:
    print(f"[red]Error: {message}[/red]")


def console_warning(message: str) -> None:
    print(f"[yellow]Warining: {message}[/yellow]")


def console_success(message: str) -> None:
    print(f"[green]{message}[/green]")



from base64 import b64encode, b64decode
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import typer
from requests.exceptions import HTTPError
from rich import print
from ruamel.yaml import YAML

from anylearn.sdk.console import (
    console_error,
    console_success,
)
from anylearn.sdk.context import get_context
from anylearn.utils import DEFAULT_ANYLEARN_HOST, get_config_path
from anylearn.utils.api import raise_for_status
from anylearn.utils.retry import RequestRetrySession


@dataclass
class Auth:
    host: str
    username: str
    access_token: str
    refresh_token: Optional[str] = None
    user_id: Optional[str] = None


__auth__: Auth = None


def configure_auth(force: bool = False) -> Optional[Auth]:
    global __auth__
    ctx = get_context()
    if not __auth__ or force:
        __auth__ = authenticate(host=ctx.host)
    return __auth__


def authenticate(host: str = DEFAULT_ANYLEARN_HOST) -> Optional[Auth]:
    """
    Try to ensure that user is authenticated to a given Anylearn host:
    1. Try to load auths from local config file.
    2. If no auths found or forced to relogin, prompt user to login.
    3. If auths found, try to login by tokens.
    4. If login bu token failed, prompt user to login.
    5. If login succeeded, dump auths to local config file.
    """
    if not host:
        host = DEFAULT_ANYLEARN_HOST
    host = host.rstrip("/")
    auths = _load()
    if not isinstance(auths, dict) or host not in auths:
        auth = None
    else:
        auth = auths[host]
        auth = _login_by_tokens(
            host=host,
            username=auth.username,
            access_token=auth.access_token,
            refresh_token=auth.refresh_token,
        )
    if not auth:
        auth = _prompt_login(host)
    _dump(auth)
    return auth


def disauthenticate(host: str = DEFAULT_ANYLEARN_HOST) -> None:
    """
    Remove auths of a given Anylearn host from local config file.
    """
    if not host:
        host = DEFAULT_ANYLEARN_HOST
    host = host.rstrip("/")
    _del(host)


def _load() -> Optional[Dict[str, Auth]]:
    config_path = get_config_path()
    if not config_path.exists():
        return None
    yaml = YAML()
    config = yaml.load(config_path)
    if not config or 'auths' not in config:
        return None
    auths = {}
    for host, auth in config['auths'].items():
        if not host or not isinstance(auth, dict):
            continue
        if not all([
            auth.get('username'),
            auth.get('access_token'),
            auth.get('refresh_token'),
            auth.get('user_id'),
        ]):
            continue
        auths[host] = Auth(
            host=host,
            username=auth.get('username'),
            access_token=b64decode(auth.get('access_token').encode()).decode(),
            refresh_token=b64decode(auth.get('refresh_token').encode()).decode(),
            user_id=auth.get('user_id'),
        )
    return auths


def _dump(auths: Union[Dict[str, Auth], Auth]) -> None:
    if auths is None:
        return
    if isinstance(auths, Auth):
        auths = {auths.host: auths}
    config_path = get_config_path()
    if not config_path.exists():
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.touch()
    yaml = YAML()
    config = yaml.load(config_path)
    if not config:
        config = {}
    if 'auths' not in config:
        config['auths'] = {}
    for _, auth in auths.items():
        if not all([
            auth.host,
            auth.username,
            auth.access_token,
            auth.refresh_token,
            auth.user_id,
        ]):
            continue
        config['auths'][auth.host] = {
            'username': auth.username,
            'access_token': b64encode(auth.access_token.encode()).decode(),
            'refresh_token': b64encode(auth.refresh_token.encode()).decode(),
            'user_id': auth.user_id,
        }
    yaml.dump(config, config_path)


def _del(hosts: Union[List[str], str]) -> None:
    config_path = get_config_path()
    if not config_path.exists():
        return None
    yaml = YAML()
    config = yaml.load(config_path)
    if not config or 'auths' not in config:
        return None
    if isinstance(hosts, str):
        hosts = [hosts]
    for host in hosts:
        if host in config['auths']:
            del config['auths'][host]
    yaml.dump(config, config_path)


def _login_by_tokens(
    host: str,
    username: str,
    access_token: str,
    refresh_token: str,
) -> Optional[Auth]:
    try:
        auth = _login_by_access_token(
            host=host,
            username=username,
            access_token=access_token,
        )
        auth.refresh_token = refresh_token
        return auth
    except HTTPError as e:
        if e.response.status_code != 401:
            return None
        return _relogin_by_refresh_token(
            host=host,
            username=username,
            refresh_token=refresh_token,
        )


def _login_by_access_token(
    host: str,
    username: str,
    access_token: str,
) -> Optional[Auth]:
    with RequestRetrySession() as sess:
        res = sess.get(
            f"{host}/api/user/me",
            headers={'Authorization': f"Bearer {access_token}"},
        )
        raise_for_status(res)
        res.encoding = "utf-8"
        data = res.json()
        if (
            'id' not in data or
            'username' not in data or
            data['username'] != username
        ):
            raise ValueError
        return Auth(
            host=host,
            username=username,
            access_token=access_token,
            user_id=data['id'],
        )


def _relogin_by_refresh_token(
    host: str,
    username: str,
    refresh_token: str,
) -> Optional[Auth]:
    with RequestRetrySession() as sess:
        try:
            res = sess.get(
                f"{host}/api/user/refresh_token",
                headers={'Authorization': f"Bearer {refresh_token}"},
            )
            raise_for_status(res)
            res.encoding = "utf-8"
            data = res.json()
            if 'token' not in data:
                raise ValueError
            auth = _login_by_access_token(
                host=host,
                username=username,
                access_token=data['token'],
            )
            auth.refresh_token = refresh_token
            return auth
        except:
            return None


def _prompt_login(host: str) -> Optional[Auth]:
    print(f"Login to {host} :")
    username = typer.prompt("Username")
    password = typer.prompt("Password", hide_input=True)
    auth = _login_by_password(
        host=host,
        username=username,
        password=password,
    )
    if auth:
        console_success(f"Login to {host} as {auth.username} succeeded.")
    else:
        console_error("Login failed. Please check your username and password.")
    return auth


def _login_by_password(
    host: str,
    username: str,
    password: str,
) -> Optional[Auth]:
    with RequestRetrySession() as sess:
        try:
            res = sess.post(
                f"{host}/api/user/login",
                data={
                    'username': username,
                    'password': password,
                },
            )
            raise_for_status(res)
            res.encoding = "utf-8"
            data = res.json()
            if not all([
                data.get('id'),
                data.get('token'),
                data.get('refresh_token'),
            ]):
                raise ValueError
            return Auth(
                host=host,
                username=username,
                access_token=data['token'],
                refresh_token=data['refresh_token'],
                user_id=data['id'],
            )
        except:
            return None



import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter


class RequestRetrySession(object):
    def __init__(
        self,
        retries=3,
        backoff_factor=1,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["HEAD", "GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "TRACE"],
        session=None,
    ):
        self.session = session or requests.session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            allowed_methods=allowed_methods,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)

    def __enter__(self):
        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()
